#include <Eigen/Core>

#include <Eigen/LU> // required for MatrixBase::determinant
#include <Eigen/SVD> // required for SVD

namespace Eigen
{
  // These helpers are required since it allows to use mixed types as parameters
  // for the Umeyama. The problem with mixed parameters is that the return type
  // cannot trivially be deduced when float and double types are mixed.
  namespace
  {
    // Always returns the type with the highest precision.
    template<typename MatrixType, typename OtherMatrixType> 
    struct MatrixTypePromotionTraits
    {
      typedef typename ei_meta_if< 
        (sizeof(ei_traits<MatrixType>::Scalar)>sizeof(ei_traits<OtherMatrixType>::Scalar)), 
        typename ei_traits<MatrixType>::Scalar, 
        typename ei_traits<OtherMatrixType>::Scalar 
      >::ret Scalar;
    };

    // Compile time return type deduction for different MatrixBase types.
    template<typename MatrixType, typename OtherMatrixType>
    struct UmeyamaTraits
    {
      enum {
        RowsAtCompileTime = EIGEN_ENUM_MIN(MatrixType::RowsAtCompileTime+1, Dynamic),
        MaxRowsAtCompileTime = EIGEN_ENUM_MIN(MatrixType::MaxRowsAtCompileTime+1, Dynamic)
      };

      typedef Matrix<typename MatrixTypePromotionTraits<MatrixType, OtherMatrixType>::Scalar,
        RowsAtCompileTime,
        RowsAtCompileTime,
        AutoAlign | (ei_traits<MatrixType>::Flags & RowMajorBit ? RowMajor : ColMajor),
        MaxRowsAtCompileTime, 
        MatrixType::MaxColsAtCompileTime
      > TransformationMatrixType;
    };
  }

  /**
  * \brief Returns the transformation between two point patterns.
  *
  * The algorithm is based on:
  * [1] "Least-squares estimation of transformation parameters between two point patterns",
  * Shinji Umeyama, PAMI 1991, DOI: 10.1109/34.88573
  *
  * It estimates parameters \f$ c, R, \f$ and \f$ t \f$ such that
  * \f[
  *   \frac{1}{n} \sum_{i=1}^n \vert\vert y_i - (cRx_i + t)^2 \vert\vert^2
  * \f]
  * is minimized.
  *
  * \param src Source points \f$ x_i \f$.
  * \param dst Destination points \f$ y_i \f$.
  * \param with_scaling Sets c=1 when <code>false</code> is passed.
  * \return The transformation \f$ [cR \vert t] \f$.
  *
  * \todo Are mixed types required / desired?
  * \todo Does MatrixBase::cast<T>() become a NOP when T == MatrixBase::Scalar?
  *       In case it does not incur additional cost, mixed types seem to be fine.
  */
  template <typename Derived, typename OtherDerived>
  typename UmeyamaTraits<Derived, OtherDerived>::TransformationMatrixType
  umeyama(const MatrixBase<Derived>& src, const MatrixBase<OtherDerived>& dst, MatrixXd& R_gt, bool with_scaling = true)
  {
    enum { Dimension = MatrixBase<Derived>::RowsAtCompileTime };

    typedef typename MatrixTypePromotionTraits<Derived, OtherDerived>::Scalar Scalar;
    
    typedef Matrix<Scalar, Dimension, 1> VectorType;
    typedef Matrix<Scalar, Dimension, Dimension> MatrixType;
    typedef typename UmeyamaTraits<Derived, OtherDerived>::TransformationMatrixType TransformationMatrixType;

    const int m = src.rows(); // dimension
    const int n = src.cols(); // number of measurements

    // required for demeaning ...
    const Scalar one_over_n = 1 / static_cast<Scalar>(n);

    // computation of mean
    VectorType src_mean = VectorType::Zero(m);
    VectorType dst_mean = VectorType::Zero(m);
    for (int i=0; i<n; ++i)
    {
      src_mean += src.col(i).cast<Scalar>();
      dst_mean += dst.col(i).cast<Scalar>();
    }
    src_mean *= one_over_n;
    dst_mean *= one_over_n;

    // demeaning of src and dst points
    MatrixType src_demean(m,n);
    MatrixType dst_demean(m,n);
    for (int i=0; i<n; ++i)
    {
      src_demean.col(i) = src.col(i).cast<Scalar>() - src_mean;
      dst_demean.col(i) = dst.col(i).cast<Scalar>() - dst_mean;
    }

    // Eq. (36)-(37)
    Scalar src_var = 0;
    Scalar dst_var = 0;
    for (int i=0; i<m; ++i)
    {
      src_var += src_demean.row(i).dot(src_demean.row(i));
      dst_var += dst_demean.row(i).dot(dst_demean.row(i));
    }
    src_var *= one_over_n;
    dst_var *= one_over_n;

    // Eq. (38)
    MatrixType sigma = dst_demean*src_demean.transpose();
    sigma *= one_over_n;

    SVD<MatrixType> svd(sigma);

    // Initialize the resulting transformation with an identity matrix...
    TransformationMatrixType Rt = TransformationMatrixType::Identity(m+1,m+1);

    // Eq. (39)
    VectorType S = VectorType::Ones(m);
    if (sigma.determinant()<0) S(m-1) = -1;

    // Eq. (40) and (43)
    const VectorType& d = svd.singularValues();
    int rank = 0; for (int i=0; i<m; ++i) if (d(i)>0.0) ++rank;
    if (rank == m-1) {
      if ( svd.matrixU().determinant() * svd.matrixV().determinant() > 0 ) {
        Rt.block(0,0,m,m) = svd.matrixU()*svd.matrixV().transpose();
      } else {
        const Scalar s = S(m-1); S(m-1) = -1;
        Rt.block(0,0,m,m) = svd.matrixU() * S.asDiagonal() * svd.matrixV().transpose();
        S(m-1) = s;
      }
    } else {
      Rt.block(0,0,m,m) = svd.matrixU() * S.asDiagonal() * svd.matrixV().transpose();
    }

    // Eq. (42)
    const Scalar c = 1/src_var * svd.singularValues().dot(S);

    // Eq. (41)
    Rt.col(m).segment(0,m) = dst_mean - c*Rt.block(0,0,m,m)*src_mean;

    if (with_scaling) Rt.block(0,0,m,m) *= c;

    return Rt;
  }
}
