#include <Eigen/Core>
#include <Eigen/Array>

#include <fstream>

#include "Umeyama.h"

using namespace Eigen;

//  Constructs a random matrix from the orthogonal group O(size).
template <typename T>
Eigen::Matrix<T, Eigen::Dynamic, Eigen::Dynamic> randMatrixO(int size)
{
  typedef Eigen::Matrix<T, Eigen::Dynamic, Eigen::Dynamic> MatrixType;

  MatrixType Q;

  int max_tries = 20;
  T error = std::numeric_limits<T>::max();

  while (error > 2*std::numeric_limits<T>::epsilon() && max_tries > 0)
  {
	  // initialize random matrix
    Q = MatrixType::Random(size, size);

	  // orthogonalize columns using the Gram-Schmidt algorithm
	  for (int col = 0; col < size; ++col)
	  {
      MatrixType::ColXpr colVec = Q.col(col);
		  for (int prevCol = 0; prevCol < col; ++prevCol)
		  {
			  MatrixType::ColXpr prevColVec = Q.col(prevCol);
        colVec -= colVec.dot(prevColVec)*prevColVec;
		  }
		  Q.col(col) = colVec.normalized();
	  }

	  // orthogonalize rows using the Gram-Schmidt algorithm
	  // this additional orthogonalization is not necessary in theory but should enhance 
    // the numerical orthogonality of the matrix
	  for (int row = 0; row < size; ++row)
	  {
      MatrixType::RowXpr rowVec = Q.row(row);
		  for (int prevRow = 0; prevRow < row; ++prevRow)
		  {
			  MatrixType::RowXpr prevRowVec = Q.row(prevRow);
			  rowVec -= rowVec.dot(prevRowVec)*prevRowVec;
		  }
		  Q.row(row) = rowVec.normalized();
	  }

	  // final check
    const T diagonal = (Q.transpose()*Q).trace();
    error = diagonal - size;
    --max_tries;
  }

  if (max_tries == 0)
    throw std::runtime_error("randMatrixO: Could not construct orthogonal matrix!");

	return Q;
}

//  Constructs a random matrix from the special orthogonal group SO(size).
template <typename T>
Eigen::Matrix<T, Eigen::Dynamic, Eigen::Dynamic> randMatrixSO(int size)
{
  typedef Eigen::Matrix<T, Eigen::Dynamic, Eigen::Dynamic> MatrixType;

	// initialize orthogonal matrix
	MatrixType Q = randMatrixO<T>(size);

	// correct for reflections, we only want rotations
	const T det_value = Q.determinant();
	if (det_value < 0) Q.row(0) *= -1;

	// final check
	if (std::abs(Q.determinant() - T(1)) > 4*std::numeric_limits<T>::epsilon()) {
		throw std::runtime_error("randMatrixSO: Could not construct special orthogonal matrix!");
	}

	return Q;
}

template <typename MatrixType, typename OtherMatrixType>
int run_test(int dim, int num_elements)
{
  typedef typename ei_traits<MatrixType>::Scalar Scalar;
  typedef typename ei_traits<OtherMatrixType>::Scalar OtherScalar;

  typedef typename ei_meta_if< 
    (sizeof(ei_traits<MatrixType>::Scalar)<sizeof(ei_traits<OtherMatrixType>::Scalar)), 
    typename ei_traits<MatrixType>::Scalar, 
    typename ei_traits<OtherMatrixType>::Scalar 
  >::ret LowPrecisionScalar;

  // MUST be positive because in any other case det(cR_t) may become negative for
  // odd dimensions!
  const double c = std::fabs(ei_random<double>());

  MatrixXd R = randMatrixSO<double>(dim);
  VectorXd t = 50*VectorXd::Random(dim,1);

  MatrixXd cR_t = MatrixXd::Identity(dim+1,dim+1);
  cR_t.block(0,0,dim,dim) = c*R;
  cR_t.block(0,dim,dim,1) = t;

  MatrixType src = MatrixType::Random(dim+1, num_elements);
  src.row(dim) = Matrix<Scalar, 1, Dynamic>::Constant(num_elements, Scalar(1));

  OtherMatrixType dst = (cR_t*src.cast<double>()).cast<OtherScalar>();

  MatrixXd cR_t_umeyama = umeyama(src.block(0,0,dim,num_elements), dst.block(0,0,dim,num_elements), R).cast<double>();

  const double error = ( cR_t_umeyama*src.cast<double>() - dst.cast<double>() ).cwise().square().sum();

  if (error > 10*std::numeric_limits<LowPrecisionScalar>::epsilon())
  {
    std::cout << "(expected, measured): " << 2*std::numeric_limits<LowPrecisionScalar>::epsilon() << ", " << error << std::endl;
    return EXIT_FAILURE;
  }

  return EXIT_SUCCESS;
}

int main(int argc, char* argv[])
{
  const int num_elements = 100;

  // works also for dimensions bigger than 3...
  for (int dim=2; dim<6; ++dim)
  {
    for (int i=0; i<100; ++i)
    {
      int ret = run_test< MatrixXd, MatrixXd >(dim, num_elements);
      if (ret != EXIT_SUCCESS) return EXIT_FAILURE;

      ret = run_test< MatrixXf, MatrixXf >(dim, num_elements);
      if (ret != EXIT_SUCCESS) return EXIT_FAILURE;

      ret = run_test< MatrixXd, MatrixXf >(dim, num_elements);
      if (ret != EXIT_SUCCESS) return EXIT_FAILURE;

      ret = run_test< MatrixXf, MatrixXd >(dim, num_elements);
      if (ret != EXIT_SUCCESS) return EXIT_FAILURE;
    }
  }

  return EXIT_SUCCESS;
}
